/*
 * Copyright (C) 1997 Brad Rossiter
 * Author: Brad Rossiter rossib@cloudnet.com
 *                       rossiter@eeyore.stcloudstate.edu
 *                       Webpage www.cloudnet.com/~rossib  
 *
 * libcgi Version 1.1 8/21/97
 * This is libcgi, C library for writing CGI applications.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA  
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <sys/types.h>
#include <unistd.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h> 

#include <stdarg.h>

#include "libcgi.h"

/*************************************************************************/
/* cgi_initialize_form()                                                 */
/*                                                                       */
/* This function is called by the CGI client program to initialize the   */
/* form.  This is a new function that replaces having to manually call   */
/* cgi_get_globals(), cgi_get_form() and cgi_print_header() and is       */
/* essentially the bulk of the version 1.1 update along with patching    */
/* other functions to abide by this newer setup.                         */
/* This will make client programming a bit easier.                       */
/*************************************************************************/

void cgi_initialize_form(HTML_FORM *form, int choice)
{
  *form = (HTML_FORM) malloc(sizeof(struct html_form));
  (*form)->globals = (cgi_gs *) malloc(sizeof(cgi_gs));

  cgi_get_globals((*form)->globals);
  cgi_print_header(choice);
  cgi_get_form((*(*form)->globals),(&(*form)->f_head),(&(*form)->f_tail));
}

/*************************************************************************/
/* cgi_get_globals()                                                     */
/*                                                                       */
/* This function needs to be called to get all the server global         */
/* environmental variables.  Most important is the CONTENT_LENGTH        */
/* This is used to take in standard input after the form is submitted    */
/* and correctly allocate memory for the variable list.                  */
/* cgi_get_globals is no longer required to be called by hand, it is now */
/* initialized in cgi_initialize_form()                                  */
/*************************************************************************/

void cgi_get_globals(cgi_gs *global)
{

/* 
 * If getenv() can't return a certian environmental variable 
 * it will return a null pointer. So these values may be null
 * the calling CGI program must check for this.
 */

  global->SERVER_SOFTWARE 	= getenv("SERVER_SOFTWARE");
  global->SERVER_NAME		= getenv("SERVER_NAME");
  global->GATEWAY_INTERFACE	= getenv("GATEWAY_INTERFACE");
  global->SERVER_PROTOCOL	= getenv("SERVER_PROTOCOL");
  global->SERVER_PORT		= getenv("SERVER_PORT");
  global->REQUEST_METHOD	= getenv("REQUEST_METHOD");
  global->PATH_INFO		= getenv("PATH_INFO");
  global->PATH_TRANSLATED	= getenv("PATH_TRANSLATED");
  global->SCRIPT_NAME		= getenv("SCRIPT_NAME");
  global->QUERY_STRING		= getenv("QUERY_STRING");
  global->REMOTE_HOST		= getenv("REMOTE_HOST");
  global->REMOTE_ADDR		= getenv("REMOTE_ADDR");
  global->AUTH_TYPE		= getenv("AUTH_TYPE");
  global->REMOTE_USER		= getenv("REMOTE_USER");
  global->REMOTE_IDENT		= getenv("REMOTE_IDENT");
  global->CONTENT_TYPE		= getenv("CONTENT_TYPE");
  global->CONTENT_LENGTH	= strtol(getenv("CONTENT_LENGTH"),'\0',10);
  global->HTTP_ACCEPT		= getenv("HTTP_ACCEPT");
  global->HTTP_USER_AGENT	= getenv("HTTP_USER_AGENT");

}

/*************************************************************************/
/* cgi_get_form()                                                        */
/*                                                                       */
/* This function is meat of the program as for as form variable          */
/* parsing is concerned.  This function takes a head and tail of         */
/* a cgi form linked list.  This list will contain all the variable      */
/* names from the html form with their coorisponding data values that    */
/* the user entered in.   This function uses all dynamic memory to       */
/* the list elements so no memory is wasted.  I do however allocate      */
/* memory for temporary usage.  This memory is freed up when it is       */
/* finished parsing.  This function is quite nasty and has some debug    */
/* code in it that I left included.  If you define DEBUG somewhere       */
/* (either in the libcgi.h or through -DDEBUG in the CFLAGS int the      */
/* Makefile then you can see how everthing gets parsed when the form is  */
/* read in.                                                              */
/*************************************************************************/

void cgi_get_form(cgi_gs global,LINK *f_head,LINK *f_tail)
{
  static char *string = NULL;

  char *ptr,
       **parse;

  int c,
      chars_read = 0,
      parse_elements = 0;

  LINK new_node;

  if(!string)
  {
	string = (char *) malloc (global.CONTENT_LENGTH + 1);
	memset(string,'\0',global.CONTENT_LENGTH + 1);

	ptr = string;

	while((chars_read < global.CONTENT_LENGTH)
		&&((c = fgetc(stdin)) != EOF))
	{
		chars_read++;
		*ptr++ = c;
	}
  }

#ifdef DEBUG
printf("Chars Read is: [%d]\nString is: [%s]\n",chars_read,string); 
#endif

  ptr = strtok(string,"&");
  parse = (char **) calloc (MAX_PARSE_VARIABLES, sizeof (char *));

  while((ptr) && (parse_elements < MAX_PARSE_VARIABLES))
  {
	parse[parse_elements] = (char *) malloc (strlen(ptr) + 1);
	strcpy(parse[parse_elements++],ptr);
	ptr = strtok(NULL,"&");
  }		

#ifdef DEBUG
	printf("Parse Elements is [%d]\n",parse_elements);
	for(c = 0; c < parse_elements; c++)
	{
		printf("parse[%d] = [%s]\n<BR>",c,parse[c]);
	}
#endif

  create_list(f_head,f_tail);

  for(c = 0; c < parse_elements; c++)
  {
	ptr = strtok(parse[c],"=");
	filter_string(ptr);

#ifdef DEBUG
printf("<BR><BR>ptr = [%s]\n<BR>",ptr);  
#endif

	new_node = (LINK) malloc(sizeof(FORM_ELEMENT));

	new_node->var_name = (char *) malloc(strlen(ptr) + 1);
	strcpy(new_node->var_name,ptr);

	ptr += strlen(ptr) + 1;

#ifdef DEBUG
printf("<BR><BR>ptr2 = [%s]\n<BR>",ptr); 
#endif
		/*
	         *	Dynamic Memory Allocation for
	         *	data elements of new nodes is
	         *	created in cgi_build_string.
		 */

	if(strlen(ptr) < 1)
	  new_node->data = NULL;
	else
	  new_node->data = cgi_build_string(ptr);

#ifdef DEBUG
printf("<BR><BR>data = [%s]\n<BR>",new_node->data); 
#endif

	insert_element(new_node,*f_tail);
  }

  free(parse);
}

/*************************************************************************/
/* create_list()                                                         */
/*                                                                       */
/* This function is used by cgi_get_form() to create the list            */
/* head and tail.  A Double Linked list is used because I feel           */
/* they are easier to work with espically when doing a lot of            */
/* traversing though lists.  The list may take a bit of                  */
/* additional memory, but the functions don't have to setup as           */
/* many local temp variables to keep track of position when              */
/* traversing.  Note:  The data elements of our head and tail            */
/* are set to NULL.  This helps cgi_getvar() return more                 */
/* gracfully when it does not find a variable a CGI client               */
/* thinks exists.                                                        */
/*************************************************************************/

void create_list(LINK *first, LINK *last)
{
  *first = (LINK) malloc(sizeof(FORM_ELEMENT));
  *last  = (LINK) malloc(sizeof(FORM_ELEMENT));

  (*first)->next  = *last;
  (*first)->prior = NULL;
  (*last)->next   = NULL;
  (*last)->prior  = *first;

  (*first)->data  = NULL;
  (*last)->data	= NULL;
}

/*************************************************************************/
/* cgi_free_form()                                                       */
/*                                                                       */
/* This function is called by the Client CGI program to free             */
/* the allocated memory associated with the list of form                 */
/* variables created in cgi_get_form().                                  */
/*************************************************************************/

void cgi_free_form(HTML_FORM *form)
{
  LINK traverse = (*form)->f_head->next;

  while(traverse != (*form)->f_tail)
  {
    free(traverse->prior);
    traverse = traverse->next;
  }

  free((*form)->f_tail);
  free((*form)->globals);
  free(*form);
}

/*************************************************************************/
/* insert_element()                                                      */
/*                                                                       */
/* This function inserts new nodes to the END of the list so I           */
/* can keep the order of the variables I set in the html form            */
/*************************************************************************/

void insert_element(LINK new_node, LINK last)
{
  new_node->next = last; 
  new_node->prior = last->prior;
  last->prior->next = new_node;
  last->prior = new_node;
}

/*************************************************************************/
/*  get_formvar()                                                        */
/*                                                                       */
/* This function traverses though the list until it either               */
/* finds the element and returns it or it fails to find the              */
/* element(CGI Variable) and returns NULL.  It returns NULL              */
/* because when it cannot find the element the traverse                  */
/* pointer is pointing to 'f_tail'.  'fm_tail' has its 'data' set        */
/* to NULL in create_list()                                              */
/*************************************************************************/

FORM_DATA get_formvar(char *var_name, HTML_FORM form)
{
  LINK traverse = form->f_head->next;

  while((traverse != form->f_tail) && 
        (strcmp(traverse->var_name,var_name) != 0))
    traverse = traverse->next;

  return(traverse->data);
}

/*************************************************************************/
/* display_list()                                                        */
/*                                                                       */
/* This function is used for testing the list for validity.              */
/* The Client CGI may want to make sure all the form                     */
/* variables are are correctly getting into the list.  But               */
/* I really only used this for developing.                               */
/*************************************************************************/

void display_list(HTML_FORM form)
{
  LINK traverse;

  traverse = form->f_head->next;

  printf("\nLinked List Informtion (Decoded Data from FORM)\n");

  while(traverse != form->f_tail)
  {
    printf("Var Name: [%s]\nVar Data: [%s]\n"
          ,traverse->var_name,traverse->data);
    traverse = traverse->next;
  }
}

/*************************************************************************/
/* hex_to_char()                                                         */
/*                                                                       */
/* This function converts the hex codes that the form                    */
/* sends through stdin into the correct text character                   */
/* and is called from the cgi_builde_string() function.                  */
/* Most of this was borrowed from NCSA's untils.c                        */
/*************************************************************************/

char hex_to_char (char *string)
{
  register char digit;

  digit = (string[0] >= 'A' ? ((string[0] & 0xdf) - 'A')
	  + 10 : (string[0] - '0'));

  digit <<=  3; /* Multiply by 16 */

  digit += (string[1] >= 'A' ? ((string[1] &0xdf) - 'A')
	  + 10 : (string[1] - '0'));

  return(digit);
} 

/*************************************************************************/
/* cgi_build_string()                                                   */
/*                                                                       */
/* This function decodes the form stdin.  It changes "+"'s               */
/* to spaces or converts the hex values to printable                     */
/* characters by calling het_to_char().                                  */
/*************************************************************************/

FORM_DATA cgi_build_string(char *raw_string)
{
  FORM_DATA string;
  int i,
      j;

  string = (FORM_DATA) malloc (strlen(raw_string) + 1);

  for(i=0,j=0;raw_string[i] != '\0';i++,j++)
  {
	if(raw_string[i] == '+')
	{
		string[j] = ' ';
	}
	else if(raw_string[i] == '%')
	{
		string[j] = hex_to_char(raw_string + i + 1);
		i += 2;
	}
	else
		string[j] = raw_string[i];
  }

  string[j] = '\0';
  filter_string(string); /* Try in weed out garbage */

  return(string);
}		 

/*************************************************************************/
/* cgi_SMTP_open()                                                       */
/*                                                                       */
/* This function opens up the SMTP port that is passed                   */
/* as a parameter from the Client CGI program.  It also                  */
/* gets the SMTP header information and sets the from,                   */
/* to and subject then places it into DATA mode so                       */
/* socket_printf() can send data strings into the message                */
/* content of the Mail-Message.  This function also                      */
/* returns the socket descriptor so that other                           */
/* functions (socket_printf() & cgi_SMTP_close()) can                    */
/* access the socket.                                                    */
/*************************************************************************/

int cgi_SMTP_open(char *smtp_host, char *mail_from
		 ,char *mail_to, char *subject)
{
  int sock, 
      bytes_rec;

  struct sockaddr_in socket_address;
  struct hostent *host_entry;
  struct servent *server_entry;

  char sendmail_header[500],
       data[1024];

  if (mail_to == NULL)
    return -1;    /* No recipient */

  /*
   *  Open endpoint for communications
   *  Using ARPA TCP Internetwork Protocol and
   *  a two-way byte stream.
   *
   */

  if ((sock = socket(AF_INET,SOCK_STREAM,0)) < 0)
  {
	return -2;   /*
                      * Error Occurred in Socket Creation     
                      * socket() should return a -1 if error  
                      * else return a descriptor value        
                      * referencing the socket.               
                      *
                      */
  }

  host_entry = gethostbyname(smtp_host);
  if (!host_entry)
  {
	close(sock);
	return -3;  /* Can't find SMTP Host */
  }

  socket_address.sin_family = AF_INET;
  memcpy ((char *)&socket_address.sin_addr,
         (char *)host_entry->h_addr,host_entry->h_length);

  if((server_entry = getservbyname("smtp","tcp")) == NULL)
  {
    /*
     * Convert Between Network byte order
     * and Host byte order.
     *
     */
	socket_address.sin_port = htons (25);
  }
  else
    socket_address.sin_port = server_entry->s_port;

  /*
   * Now try to initiate a connection on the
   * created socket.  This function reuturns 0
   * is binding to the socket is successfull. 
   * Otherwise a -1 is returned.
   *
   */

  if(connect(sock,(struct sockaddr *)&socket_address,
	sizeof(socket_address)))
  {
    close(sock);
    return -4;
  }

  /* 
   * Now we are going to start receiving messages
   * from the socket.  recv() returns the number
   * of bytes received, or -1 if there is an error.
   *
   */

  if((bytes_rec = recv(sock,sendmail_header,sizeof(sendmail_header),0)) == -1)
  {
    close(sock);
    return -5;
  }

/* Required for NCR Unix */
#ifdef NCR
  if((bytes_rec = recv(sock,sendmail_header,sizeof(sendmail_header),0)) == -1)
  {
    close(sock);
    return -5;
  }
#endif

  sprintf(data,"MAIL FROM: %s\r\n",mail_from);
  if(send(sock,data,strlen(data),0) <= 0)
  {
    close(sock);
    return -6;
  }
  
  if((bytes_rec = recv(sock,data,300,0)) <= 0)
  {
    close(sock);
    return -7;
  }
  
  if(strncmp(data,SMTP_OK,strlen(SMTP_OK)) != 0)
  {
    close(sock);
    return -8; 
  }

  sprintf(data,"RCPT TO: %s\r\n",mail_to);
  if(send(sock,data,strlen(data),0) <= 0)
  {
    close(sock);
    return -9;
  }

  if((bytes_rec = recv(sock,data,300,0)) <= 0)
  {
    close(sock);
    return -10;
  }
  
  if(strncmp(data,SMTP_OK,strlen(SMTP_OK)) !=0)
  {
    close(sock);
    return -11;
  }
  
  strcpy(data,"DATA\r\n");
  if(send(sock,data,strlen(data),0) <= 0)
  {
    close(sock);
    return -12;
  }
  
  if((bytes_rec = recv(sock,data,300,0)) <= 0)
  {
    close(sock);
    return -13;
  }

  if(strncmp(data,"354 ",4) != 0)
  {
    close(sock);
    return -14;
  }

  socket_printf(sock,"Subject: %s\n",subject); 

  return sock;
}

/*************************************************************************/
/* cgi_SMTP_close()                                                      */
/*                                                                       */
/* This function uses the socket descriptor 'sock' passed from           */
/* the Client CGI program to close the socket and also ENDS              */
/* the DATA section of a mail message.                                   */
/*************************************************************************/

int cgi_SMTP_close(int sock)
{
  char data[300];
  int bytes_rec;

  strcpy(data,"\r\n.\r\n");

  if(send(sock,data,strlen(data),0) <= 0)
  {
    close(sock);
    return -2;
  }

  if((bytes_rec = (recv(sock,data,300,0))) <= 0)
  {
    close(sock);
    return -3;
  }

  if(strncmp(data,SMTP_OK,strlen(SMTP_OK)) != 0)
  {
    close(sock);
    return -4;
  }

  strcpy(data,"QUIT\r\n");
  send(sock,data,strlen(data),0);

  return(close(sock));

}

/*************************************************************************/
/* socket_printf()                                                       */
/*                                                                       */
/* This is a really neet function that I thought should help             */
/* the CGI Client's formating to E-mail much easier and more             */
/* standard to C.  socket_printf() works exactally (well, I              */
/* think) like sprintf() except instead of giving a FILE                 */
/* descriptor as the first parameter, you give a socket                  */
/* descriptor.  This 'sock' should be passed back from                   */
/* cgi_SMTP_open() and then used in socket_printf().                     */
/*************************************************************************/

int socket_printf(int sock, char *message,...)
{
  va_list argptr;

  int char_count;

  char message_buffer[MAX_MESSAGE_BUFFER];

  /* 
   * Max Length of ONE line or
   * message sent to socket_printf.
   * See libcgi.h for setting/changing.
   *
   */

  va_start(argptr,message);
  char_count = vsprintf(message_buffer,message,argptr);
  va_end(argptr);

  if (send(sock,message_buffer,strlen(message_buffer),0) <= 0)
  {
    close(sock);
    return -1;
  }
 
  return 0;
} 

/*************************************************************************/
/* page_bottom()                                                         */
/*                                                                       */
/* Since many sites use the exact same thing at the bottom of            */
/* every page on their web site I thought it would be nice to            */
/* easily be able to display this with one short function call           */
/* instead of a bunch of HTML coded into a printf statement in C.        */
/* page_bottom() opens the file ~cgi-bin/html/bottom.html (in libcgi.h)  */
/* for reading and then displays it to standard output which if          */
/* the Client CGI's program uses the cgi_print_header(TRUE) will         */
/* be BACK to the users screen on the web client.  The lines             */
/* parameter is used to set how far down the screen you want your        */
/* 'bottom'.                                                             */
/*************************************************************************/
 
int page_bottom(int lines)
{
  FILE *fp;
  int c;

  if((fp = fopen(BOTTEM,"r")) == NULL)
    return (-1);

  printf("<PRE>");
  for(c=0;c<lines;c++)
    printf("\n");
  printf("</PRE>");

  while ((c = fgetc(fp)) != EOF)
    printf("%c",c);

  fclose(fp);

  return 0;	
}

/*************************************************************************/
/* cgi_print_header()                                                    */
/*                                                                       */
/* This function is called from cgi_initialize_form() to set the print  */
/* header.  Passing TRUE (or 1) will tell the HTTPD server that          */
/* the cgi will be returning HTML code back to the user.  Passing        */
/* FALSE (or O) will tell the HTTPD server to expect nothing back.       */
/*************************************************************************/

void cgi_print_header(int choice)
{
  if(choice)
    printf("Content-type: text/html\n\n");
  else
    printf("HTTP/1.0 204 No Response\nContent-type: text/html\n\n");
}

/*************************************************************************/
/* filter_string()                                                       */
/*                                                                       */
/* This function does its best to filter out Control Chars.   It         */
/* is called from the cgi_get_form() function.  This is usefull          */
/* because we are never exactally sure what the user has typed in.       */
/*************************************************************************/

void filter_string(char *string)
{
  char *ptr;

  ptr = string;

  while(*ptr != '\0')
  {
    if((iscntrl(*ptr)) && (*ptr != '\n'))
      *ptr = ' ';
    ptr++;
  }
}

/************************************************************************/
